' Billardsimulation: Simulation vllig reibungsfreier, vollelastischer
' Stsse (Eta=1) als animierte, farbige Realtime-Simulation
' (c) 1996, 1997 by Andreas Meile, 8242 Hofen SH
' Erweiterte Version mit optionaler Gravitation (Schwerkraft wie bei
' einem Flipperkasten)
' Ausserdem: Halde fr Ereignisschlange

' Eigene Datentypen

TYPE Vektor   ' 2D-Vektor
  x AS SINGLE ' x-Komponente
  y AS SINGLE ' y-Komponente
END TYPE

TYPE Kugel     ' Beschreibung einer Billardkugel
  t0 AS SINGLE ' Bezugszeitpunkt t0
  p AS Vektor  ' Position zur Zeit t0
  v AS Vektor  ' momentane Geschwindigkeit
  m AS SINGLE  ' Masse
  r AS SINGLE  ' Radius
  f AS INTEGER ' Farbe
END TYPE

TYPE Bandenwand ' Beschreibung eines Bandenwandstckes
  b AS Vektor  ' Eckpunkt
  e AS Vektor  ' Einheitsvektor, gerichtet zum nchsten Eckpunkt
  e2 AS Vektor ' Transformationsvektor: Entspricht e bei komplexer Betracht.
  a AS SINGLE  ' Abstand des Koordinatenursprungs zur Bandenwand
  g AS SINGLE  ' Querkomponente Schwerkraftbeschleunigungskonstante g
END TYPE

TYPE Stosszeit ' Stosszeiteintrag
  t AS SINGLE ' Zeitpunkt
  eind1 AS INTEGER ' Index auf 1. Element
  eind2 AS INTEGER ' Index auf 2. Element
END TYPE

DECLARE SUB BewegeKugel (k AS Kugel, t!)
DECLARE SUB UmlenkungBande (bw AS Bandenwand, k AS Kugel, t!)
DECLARE SUB UmlenkungStoss (k1 AS Kugel, k2 AS Kugel, t!)
DECLARE SUB WannBandeRundum (b() AS Bandenwand, k AS Kugel, zpMax!, bnr%)
DECLARE SUB WannStoss (k1 AS Kugel, k2 AS Kugel, t!, StossJa%)
DECLARE SUB WannBande (bw AS Bandenwand, k AS Kugel, t!, StossJa%)
DECLARE SUB ZeichneKugel (k AS Kugel, n%)
DECLARE SUB StossEinfuegen (S AS Stosszeit)
DECLARE SUB LoescheEreignisse (KugNr%)
DECLARE FUNCTION Quad! (S!)

CONST nWaendeMax% = 50       ' Bande darf max. aus 50 Wnden bestehen
CONST nKugelnMax% = 50       ' Max. 50 Kugeln im Szenario
CONST SegMax% = 1000         ' Maximal erlaubte Anzahl Segmente
CONST yxVerh! = 3! / 4!      ' Bildverhltnis
CONST Unendl! = 3.402823E+38 ' wirkt praktisch als unendlich ()

DIM SHARED StEreig(0 TO nKugelnMax% * (nKugelnMax% + 1) + 1) AS Stosszeit
DIM SHARED AnzSt%, AktZeit!, Pi!, ks(1 TO SegMax%) AS Vektor, g AS Vektor
DIM SHARED ksp%(1 TO nKugelnMax% + 1), TonEin%, nKugeln%, nWaende%, Tol!
DIM Bande(1 TO nWaendeMax%) AS Bandenwand, BKugel(1 TO nKugelnMax%) AS Kugel
DIM Sto AS Stosszeit

Pi! = 4! * ATN(1!)

' Vorgabewerte vergeben

AktZeit! = -500!
FOR i% = 1 TO nKugelnMax%
  BKugel(i%).t0 = AktZeit!
  BKugel(i%).p.x = 0!
  BKugel(i%).p.y = 0!
  BKugel(i%).v.x = 0!
  BKugel(i%).v.y = 0!
  BKugel(i%).m = 100!
  BKugel(i%).r = 10!
  BKugel(i%).f = 10
NEXT i%
TonEin% = -1
Tol! = .4
VerklFakt! = 1.01  ' Verkleinerungsfaktor; wre dieser 1.0, dann wrde die
                   ' Bande direkt den Bildschrimrand berhren (Steuerung der
                   ' Grsse am Bildschirm
VerlaFakt! = 1! ' Verlangsamungsfaktor; erlaubt eine Geschwindigkeits-
                ' rcknahme bei langsameren PC's
g.x = 0!        ' Vorgabemssig keine Schwerkraft
g.y = 0!        ' => Weltall-Verhltnisse

nKugeln% = 0 ' Am Anfang weder Kugeln
nWaende% = 0 ' noch keine Wnde
KNr% = 1 ' Zeiger, welche Kugel gerade verarbeitet werden muss

' Dateiname vom Benutzer verlangen

CLS
PRINT "Simulation eines vllig reibungsfreien Billardtisches"
PRINT "(c) 1996, 1997 by Andreas Meile"
PRINT
INPUT "Name einer Textdatei mit Startszenario"; d$

' Datei einlesen und Werte fllen

OPEN d$ FOR INPUT AS 1
WHILE NOT EOF(1)
  LINE INPUT #1, a$
  t% = INSTR(a$, "%")
  IF t% THEN
    a$ = LEFT$(a$, t% - 1)   ' Kommentar abschneiden
  END IF
  t% = INSTR(a$, "=")
  IF t% THEN    ' Kommandozeile
    w$ = MID$(a$, t% + 1)  ' Werte
    SELECT CASE LCASE$(LEFT$(a$, t% - 1))
    CASE "bec" ' Bandenecke
      IF nWaende% < nWaendeMax% THEN ' Nur, falls noch Platz
        nWaende% = nWaende% + 1
        i% = INSTR(w$, " ")
        Bande(nWaende%).b.x = VAL(LEFT$(w$, i% - 1))
        Bande(nWaende%).b.y = VAL(MID$(w$, i% + 1))
        IF nWaende% = 1 THEN
          xMin! = Bande(nWaende%).b.x
          xMax! = Bande(nWaende%).b.x
          yMin! = Bande(nWaende%).b.y
          yMax! = Bande(nWaende%).b.y
        ELSE
          IF Bande(nWaende%).b.x < xMin! THEN
            xMin! = Bande(nWaende%).b.x
          END IF
          IF Bande(nWaende%).b.x > xMax! THEN
            xMax! = Bande(nWaende%).b.x
          END IF
          IF Bande(nWaende%).b.y < yMin! THEN
            yMin! = Bande(nWaende%).b.y
          END IF
          IF Bande(nWaende%).b.y > yMax! THEN
            yMax! = Bande(nWaende%).b.y
          END IF
        END IF
      END IF
    CASE "kug" ' Weitere Kugel
      i% = VAL(w$)
      IF i% >= 1 AND i% <= nKugelnMax% THEN ' Muss im gltigen Bereich liegen
        KNr% = i%
        IF i% > nKugeln% THEN
          nKugeln% = i%
        END IF
      END IF
    CASE "pos" ' Festlegung der Position
      i% = INSTR(w$, " ")
      BKugel(KNr%).p.x = VAL(LEFT$(w$, i% - 1))
      BKugel(KNr%).p.y = VAL(MID$(w$, i% + 1))
    CASE "vel" ' Geschwindigkeit
      i% = INSTR(w$, " ")
      BKugel(KNr%).v.x = VAL(LEFT$(w$, i% - 1))
      BKugel(KNr%).v.y = VAL(MID$(w$, i% + 1))
    CASE "mas" ' Masse
      BKugel(KNr%).m = VAL(w$)
    CASE "rad" ' Radius
      BKugel(KNr%).r = VAL(w$)
    CASE "dic" ' Festlegung der Masse ber die Dichte
      di! = VAL(w$)
      BKugel(KNr%).m = BKugel(KNr%).r ^ 3! * 4! * Pi! / 3! * di!
    CASE "far" ' Farbe festlegen
      BKugel(KNr%).f = VAL(w$)
    CASE "ton" ' Ton
      TonEin% = LCASE$(LEFT$(w$, 1)) = "e"
    CASE "tol" ' Zeichengenauigkeit
      Tol! = VAL(w$)
    CASE "vfa" ' Verkleinerungsfaktor zur Steuerung der Grsse am Bildschirm
      VerklFakt! = VAL(w$)
    CASE "lfa" ' Verlangsamungsfaktor zur Anpassung an langsamere PC's
      VerlaFakt! = VAL(w$)
    CASE "bes" ' Schwerkraft-Gravitationskonstante g
      i% = INSTR(w$, " ")
      g.x = VAL(LEFT$(w$, i% - 1))
      g.y = VAL(MID$(w$, i% + 1))
    CASE "nei" ' Schwerkraft als Tischneigung mit Richtung definieren
      i% = INSTR(w$, " ")
      j% = INSTR(i% + 1, w$, " ")
      nWink! = VAL(MID$(w$, i% + 1, j% - i% - 1)) * Pi! / 180!
      gBetr! = VAL(LEFT$(w$, i% - 1)) * SIN(nWink!)
      rWink! = VAL(MID$(w$, j% + 1)) * Pi! / 180!
      g.x = gBetr! * COS(rWink!)
      g.y = gBetr! * SIN(rWink!)
    END SELECT
  END IF
WEND
CLOSE 1

' Grssendaten fr Bildschirmdarstellung vorbereiten

xm! = (xMax! + xMin!) / 2! ' Mittelpunkt am Bildschirm
ym! = (yMax! + yMin!) / 2!
xb! = (xMax! - xMin!) * VerklFakt!
yb! = (yMax! - yMin!) * VerklFakt!
IF xb! * yxVerh! > yb! THEN
  yb! = xb! * yxVerh!
ELSE
  xb! = yb! / yxVerh!
END IF

' Zeichentoleranz
TolEff! = Tol! * yb! / 200! ' (200 = Auflsung EGA)

' Segmente fr Kreiszeichenvorgang vorberechnen sowie Tempoanpassung

g.x = g.x / Quad(VerlaFakt!)
g.y = g.y / Quad(VerlaFakt!)
h% = 0
FOR i% = 1 TO nKugeln%
  BKugel(i%).v.x = BKugel(i%).v.x / VerlaFakt!
  BKugel(i%).v.y = BKugel(i%).v.y / VerlaFakt!
  ksp%(i%) = h%
  a% = CINT(2! * Pi! * SQR(BKugel(i%).r * .125 / TolEff!) + .49)
  IF a% < 1 THEN
    a% = 1
  END IF
  IF h% + a% > SegMax% - 3 * (nKugeln% - i%) THEN
    a% = 3  ' Schutz vor Feldberlauf
  END IF
  sw! = 2! * Pi! / CSNG(a%)
  FOR j% = 1 TO a%
    w! = CSNG(j%) * sw!
    ks(h% + j%).x = BKugel(i%).r * COS(w!)
    ks(h% + j%).y = BKugel(i%).r * SIN(w!)
  NEXT j%
  h% = h% + a%
NEXT i%
ksp%(nKugeln% + 1) = h%

' Banden umrechnen
FOR i% = 1 TO nWaende%
  dx! = Bande(i% MOD nWaende% + 1).b.x - Bande(i%).b.x
  dy! = Bande(i% MOD nWaende% + 1).b.y - Bande(i%).b.y
  Betr! = SQR(dx! * dx! + dy! * dy!)
  Bande(i%).e.x = dx! / Betr!
  Bande(i%).e.y = dy! / Betr!
  Bande(i%).e2.x = Quad(Bande(i%).e.x) - Quad(Bande(i%).e.y)
  Bande(i%).e2.y = 2! * Bande(i%).e.x * Bande(i%).e.y
  Bande(i%).a = Bande(i%).b.x * Bande(i%).e.y - Bande(i%).b.y * Bande(i%).e.x
  Bande(i%).g = g.x * Bande(i%).e.y - g.y * Bande(i%).e.x
NEXT i%

SCREEN 7, , 4  ' Niedrige EGA-Auflsung, Speicherseite 2 mit leerem Feld
CLS

' Bildschirmausschnitt setzen
WINDOW (xm! - xb! / 2!, ym! - yb! / 2!)-(xm! + xb! / 2!, ym! + yb! / 2!)

COLOR 15
PSET (Bande(nWaende%).b.x, Bande(nWaende%).b.y)
FOR i% = 1 TO nWaende%
  LINE -(Bande(i%).b.x, Bande(i%).b.y)
NEXT i%

' Queue vorbereiten
StEreig(0).t = -Unendl!  ' gengend kleiner Wert als Elternsuchstopper
FOR i% = 1 TO UBOUND(StEreig)
  StEreig(i%).t = Unendl!   ' gengend grosser Wert als Kindsuchstopper
NEXT i%
AnzSt% = 0

' Stossereigniszeiten berechnen und in die Halde einfgen

FOR i% = 1 TO nKugeln%
  WannBandeRundum Bande(), BKugel(i%), Sto.t, nr%
  Sto.eind1 = -nr%
  Sto.eind2 = i%
  StossEinfuegen Sto
  FOR j% = i% + 1 TO nKugeln%
    WannStoss BKugel(i%), BKugel(j%), Sto.t, sj%
    IF sj% THEN
      Sto.eind1 = j%
      StossEinfuegen Sto
    END IF
  NEXT j%
NEXT i%

sc% = 0
SCREEN , , 0
tk! = TIMER - AktZeit! ' Zeitverschiebungskonstante, damit mglichst nahe bei
' 0 gerechnet wird (Rundungsfehler!)

' Ti! : Zeitgltter, um den groben Takt von nur 18.2 Clocks/Sek der
' Systemuhr auszugltten
DIM Ti!(3)
ts! = 0!
FOR i% = 0 TO 3
  Ti!(i%) = TIMER - tk!
  ts! = ts! + Ti!(i%)
NEXT i%
AktZeit! = TIMER - tk!

WHILE INKEY$ = ""
  t1! = TIMER - tk!
  AktZeit! = AktZeit! + (4! * t1! - ts!) / 10!
  ts! = ts! - Ti!(sc%) + t1!
  Ti!(sc%) = t1!
  sc% = sc% + 1 AND 3
  WHILE StEreig(1).t < AktZeit!
    e1% = StEreig(1).eind1
    e2% = StEreig(1).eind2
    IF e1% < 0 THEN
      UmlenkungBande Bande(-e1%), BKugel(e2%), StEreig(1).t
      LoescheEreignisse e2%
      WannBandeRundum Bande(), BKugel(e2%), Sto.t, h%
      Sto.eind1 = -h%
      Sto.eind2 = e2%
      StossEinfuegen Sto
      FOR i% = 1 TO nKugeln%
        IF i% <> e2% THEN
          WannStoss BKugel(e2%), BKugel(i%), Sto.t, sj%
          IF sj% THEN
            Sto.eind1 = i%
            StossEinfuegen Sto
          END IF
        END IF
      NEXT i%
    ELSE
      UmlenkungStoss BKugel(e1%), BKugel(e2%), StEreig(1).t
      LoescheEreignisse e1%
      LoescheEreignisse e2%
      FOR i% = e1% TO e2% STEP e2% - e1%
        WannBandeRundum Bande(), BKugel(i%), Sto.t, h%
        Sto.eind1 = -h%
        Sto.eind2 = i%
        StossEinfuegen Sto
        FOR j% = 1 TO nKugeln%
          IF j% <> e1% AND j% <> e2% THEN
            WannStoss BKugel(i%), BKugel(j%), Sto.t, sj%
            IF sj% THEN
              Sto.eind1 = j%
              StossEinfuegen Sto
            END IF
          END IF
        NEXT j%
      NEXT i%
    END IF
  WEND
  PCOPY 4, sc% ' Leeres Feld kopieren
  SCREEN , , sc%, sc% - 1 AND 3 ' Erst jetzt die Arbeitsseite wechseln
  FOR i% = 1 TO nKugeln%
    ZeichneKugel BKugel(i%), i%
  NEXT i%
  SCREEN , , sc% ' Sichtbar- und Arbeitsseite gleich (wegen Bug in PCOPY)
WEND

SCREEN 0
WIDTH 80, 25

' Nachfhren der Kugelbewegung
SUB BewegeKugel (k AS Kugel, t!)
  dt! = t! - k.t0
  dt2! = dt! * dt! / 2!
  k.p.x = k.p.x + k.v.x * dt! + g.x * dt2!
  k.p.y = k.p.y + k.v.y * dt! + g.y * dt2!
  k.v.x = k.v.x + g.x * dt!
  k.v.y = k.v.y + g.y * dt!
  k.t0 = t!
END SUB

SUB LoescheEreignisse (KugNr%)
  DIM a AS Stosszeit
  i% = 1
  WHILE i% <= AnzSt%
    IF StEreig(AnzSt%).eind1 = KugNr% OR StEreig(AnzSt%).eind2 = KugNr% THEN
      StEreig(AnzSt%).t = Unendl!
      AnzSt% = AnzSt% - 1
    ELSEIF StEreig(i%).eind1 = KugNr% OR StEreig(i%).eind2 = KugNr% THEN
      a = StEreig(AnzSt%)
      t% = i%
      DO
        t2% = t% \ 2
      IF StEreig(t2%).t <= a.t THEN EXIT DO
        StEreig(t%) = StEreig(t2%)   ' kleiner als die Eltern => nachrcken
        t% = t2%
      LOOP
      DO
        t2% = 2 * t%
        IF StEreig(t2% + 1).t < StEreig(t2%).t THEN
          t2% = t2% + 1   ' rechtes Kind kleiner
        END IF
      IF StEreig(t2%).t >= a.t THEN EXIT DO
        StEreig(t%) = StEreig(t2%)
        t% = t2%
      LOOP
      StEreig(t%) = a
      StEreig(AnzSt%).t = Unendl! ' gengend grosser Wert als Stopper
      AnzSt% = AnzSt% - 1
    ELSE
      i% = i% + 1
    END IF
  WEND
END SUB

' Quadrat eines Wertes bestimmen
FUNCTION Quad! (S!)
  Quad! = S! * S!
END FUNCTION

SUB StossEinfuegen (S AS Stosszeit)
  AnzSt% = AnzSt% + 1
  t% = AnzSt%
  DO
    t2% = t% \ 2
  IF StEreig(t2%).t <= S.t THEN EXIT DO
    StEreig(t%) = StEreig(t2%) ' kleiner als die Eltern => nachrcken
    t% = t2%
  LOOP
  StEreig(t%) = S
END SUB

SUB UmlenkungBande (bw AS Bandenwand, k AS Kugel, t!)
  BewegeKugel k, t!
  vx! = k.v.x
  vy! = k.v.y
  k.v.x = vx! * bw.e2.x + vy! * bw.e2.y ' Berechnung der
  k.v.y = vx! * bw.e2.y - vy! * bw.e2.x ' Abprallreflektion
  IF TonEin% THEN
    SOUND 440, .3
  END IF
END SUB

' Stattfindender Stoss anhand des Impulssatzes berechnen
SUB UmlenkungStoss (k1 AS Kugel, k2 AS Kugel, t!)
  BewegeKugel k1, t!
  BewegeKugel k2, t!
  dx! = k2.p.x - k1.p.x
  dy! = k2.p.y - k1.p.y
  Betr! = SQR(dx! * dx! + dy! * dy!)
  ex! = dx! / Betr!
  ey! = dy! / Betr!
  vu1! = k1.v.x * ex! + k1.v.y * ey!
  vv1! = k1.v.x * ey! - k1.v.y * ex!
  vu2! = k2.v.x * ex! + k2.v.y * ey!
  vv2! = k2.v.x * ey! - k2.v.y * ex!
  sm! = k1.m + k2.m
  dm! = k1.m - k2.m
  vu1a! = (2! * k2.m * vu2! + dm! * vu1!) / sm! ' Anwendung Energie-
  vu2a! = (2! * k1.m * vu1! - dm! * vu2!) / sm! ' und Impulssatz
  k1.v.x = vu1a! * ex! + vv1! * ey!
  k1.v.y = vu1a! * ey! - vv1! * ex!
  k2.v.x = vu2a! * ex! + vv2! * ey!
  k2.v.y = vu2a! * ey! - vv2! * ex!
  IF TonEin% THEN
    SOUND 880, .5
  END IF
END SUB

' Einzelne Bandenwand prfen, wann, wenn berhaupt, ein Abprall stattfindet
SUB WannBande (bw AS Bandenwand, k AS Kugel, t!, StossJa%)
  vu! = k.v.x * bw.e.y - k.v.y * bw.e.x
  pu! = k.p.x * bw.e.y - k.p.y * bw.e.x
  IF bw.g <> 0! THEN
    u! = vu! / bw.g
    Diskr! = u! * u! + 2! * (bw.a - pu! - k.r) / bw.g
    StossJa% = Diskr! > 0! ' Spezialfall Streifen noch nicht ausreichend!
    IF StossJa% THEN
      IF bw.g > 0! THEN
        t! = k.t0 - u! + SQR(Diskr!)
      ELSE
        StossJa% = vu! > 0!
        IF StossJa% THEN
          t! = k.t0 - u! - SQR(Diskr!)
        END IF
      END IF
    END IF
  ELSE
    ' Spezialfall: Keine Beschleunigung in Querrichtung
    StossJa% = vu! > 0! ' muss in Richtung Bandenwand zeigen!
    IF StossJa% THEN
      t! = k.t0 + (bw.a - k.r - pu!) / vu!
    END IF
  END IF
END SUB

SUB WannBandeRundum (b() AS Bandenwand, k AS Kugel, zpMin!, bnr%)
  zpMin! = Unendl!  ' irgend etwas >86400 (Maximum von TIMER)
  FOR i% = 1 TO nWaende%
    WannBande b(i%), k, zp!, sj%
    IF sj% AND zp! < zpMin! THEN
      zpMin! = zp!
      bnr% = i%
    END IF
  NEXT i%
END SUB

' Berechnung, wann ein evtl. Stoss zwischen zwei Kugeln stattfindet
SUB WannStoss (k1 AS Kugel, k2 AS Kugel, t!, StossJa%)
  dt! = k2.t0 - k1.t0
  ux! = g.x * dt!
  uy! = g.y * dt!
  dvx! = k1.v.x - k2.v.x + ux!
  dvy! = k1.v.y - k2.v.y + uy!
  IF dvx! = 0! AND dvy! = 0! THEN ' genau gleiche
    StossJa% = 0 ' Geschwindigkeit => kein Stoss!
    EXIT SUB
  END IF
  dt! = (k2.t0 + k1.t0) / 2!
  cx! = k1.p.x - k2.p.x - k1.v.x * k1.t0 + k2.v.x * k2.t0 - ux! * dt!
  cy! = k1.p.y - k2.p.y - k1.v.y * k1.t0 + k2.v.y * k2.t0 - uy! * dt!
  a! = k1.r + k2.r
  u! = dvx! * dvx! + dvy! * dvy!
  Diskr! = a! * a! * u! - Quad(dvx! * cy! - dvy! * cx!)
  StossJa% = Diskr! > 0!
  IF StossJa% THEN
    h! = -dvx! * cx! - dvy! * cy!
    IF k1.t0 > k2.t0 THEN
      StossJa% = h! > u! * k1.t0  ' frher stattgefundene Stsse
    ELSE                          ' knnen nicht mehr stattfinden!
      StossJa% = h! > u! * k2.t0
    END IF
  END IF
  IF StossJa% THEN
    t! = (h! - SQR(Diskr!)) / u!
  END IF
END SUB

SUB ZeichneKugel (k AS Kugel, n%)
  dt! = AktZeit! - k.t0
  dt2! = dt! * dt! / 2!
  mx! = k.p.x + dt! * k.v.x + dt2! * g.x
  my! = k.p.y + dt! * k.v.y + dt2! * g.y
  COLOR k.f
  PSET (mx! + k.r, my!)
  FOR i% = ksp%(n%) + 1 TO ksp%(n% + 1)
    LINE -(mx! + ks(i%).x, my! + ks(i%).y)
  NEXT i%
END SUB

